/*******************************************************
 * A Particle System
 *******************************************************/
#include <windows.h>
#include "glew.h"

#include <GL/glut.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
//Good ol' pi
#define PI 3.14159265358979623
//Path following spring constants
#define KS 2.0f
#define KD 0.2f
//particle emission rate (parts/unit time)
#define RATE 100
//maximum number of particles allowed.  10000 looks pretty nice
#define MAX_NUM_P 10000
//maximum particle age.  This needs to equal max/rate, or the system gets spurty
//since more are born than are dying.  This get equilibrium.
#define EUTHENIZE MAX_NUM_P/RATE
//the length of the phase space state vector.  defined, but rarely changed
#define STATE_LENGTH 6 
//max size of a whole pile of states.  defined but rarely changed
#define MAX_STATE_SIZE MAX_NUM_P*STATE_LENGTH 
//how the camera rotates
float R=0.0f;
//if changing=0, we're moving the emitter in x
//if changing=1, we're moving the emitter in y
//if changing=2, we're moving the emitter in z
//see keyboard() for more info
int changing=0;
int numParticles;
//mode variables

int gravity=0;
int pathf=0;
int dipole=0;
int floored=0;
float doExplode=0;
int whack=0;

static int window_width = 512;
static int window_height = 512;

//GLubyte checkImage[WIDTH][HEIGHT][3];
GLuint tex;
float matedif[4] = {0.0,0.0,1.0,0.6};
float matespec[4] = {1.0,0.0,1.0,1.0};
float * state;
float * stateDeriv;
/*
 * Particle System Structures
 */

typedef struct _P {
	float x[3];
	float c[4];
	float v[3];
	float f[3];
	int age;
	int tpathjoin; //when this particle joined the curve following scene
	float m;
	struct _P * dual; //a "friend" particle
} Particle; 

typedef struct {
	int n;
	int numParticles; //currently living particles
	int partRate; //current rate of emission
	float t; //TIME
	float x[3]; //position of emitter
	Particle *p;
} ParticleSystem;

ParticleSystem  * P;
Particle * parts;


//make the particle system we're using the one we have

/*
 * End of Particle System Structures
 */

inline float rad(float deg){
	return deg*(PI/180);
}

inline float norm(float ax,float ay,float az){
	return 1/sqrt(ax*ax+ay*ay+az*az);
}




/*
 * Particle System Helper Functions
 */
//This function uses the length of the phase space
//state vector and the number of living particles 
//to figure out how much space we'll need.
int psGetStateDimension(ParticleSystem *ps){
	return STATE_LENGTH*(ps->n);
}

//Set the State vector(dst) to the current state of
//the particle system
void psGetParticleState(ParticleSystem *ps, float *dst){
	int i;
	for(i=0;i<ps->numParticles;i++){
		*(dst++)=ps->p[i].x[0];
		*(dst++)=ps->p[i].x[1];
		*(dst++)=ps->p[i].x[2];
		*(dst++)=ps->p[i].v[0];
		*(dst++)=ps->p[i].v[1];
		*(dst++)=ps->p[i].v[2];
	}
}

//Set the Particle System to the updated state of
//the State Vector 
void psSetParticleState(ParticleSystem *ps, float *src){
	int i;
	for(i=0;i<ps->numParticles;i++){
		ps->p[i].x[0]=*(src++);
		ps->p[i].x[1]=*(src++);
		ps->p[i].x[2]=*(src++);
		ps->p[i].v[0]=*(src++);
		ps->p[i].v[1]=*(src++);
		ps->p[i].v[2]=*(src++);
	}
}

//This determines the derivitive of the state of the 
//P. Sys. and puts it into dst.  This is used in the 
//Euler's method solver.
void psStateDeriv(ParticleSystem *ps, float *dst){
	int i;
	for(i=0;i<ps->numParticles; i++){
		*(dst++)=ps->p[i].v[0];
		*(dst++)=ps->p[i].v[1];
		*(dst++)=ps->p[i].v[2];
		*(dst++)=ps->p[i].f[0]/ps->p[i].m;
		*(dst++)=ps->p[i].f[1]/ps->p[i].m;
		*(dst++)=ps->p[i].f[2]/ps->p[i].m;
	}
}
void pInit(ParticleSystem * ps,Particle *p){
	/*initialize with some random perturbations in V and X*/
	p->x[0] = ps->x[0]+((((float)rand())/RAND_MAX)-.5)*.1;
	p->x[1] = ps->x[1]+((((float)rand())/RAND_MAX)-.5)*.1;
	p->x[2] = ps->x[2]+((((float)rand())/RAND_MAX)-.5)*.1;
	p->v[0] = (((float)rand())/RAND_MAX)*1-(((float)rand())/RAND_MAX)*1;
	p->v[1] = 1.0f+(((float)(rand())/RAND_MAX))*3.0f;
	p->v[2] = (((float)rand())/RAND_MAX)*1-(((float)rand())/RAND_MAX)*1;
	p->f[0] = 0;
	p->f[1] = 0;
	p->f[2] = 0;
	p->c[0] = 1.0;
	p->c[1] = 0.0;
	p->c[2] = 0.0;
	p->m = 1;
	p->age = 0;
	p->tpathjoin=ps->t;
	if(p->dual->age)
		pInit(ps,p->dual);
}


void psApplyGravity(ParticleSystem * ps){
	/*Very simply applies Gravity to particles*/
	int i;
	for(i=0; i<ps->numParticles; i++){
		float ax, ay, az,an;
		ax = ps->p[i].x[0];
		ay = ps->p[i].x[1];
		az = ps->p[i].x[2];
		an=norm(ax,ay,az);
		ax=ax/an;
		ay=ay/an;
		az=az/an;
		ps->p[i].f[1]-=1.0;
	}
}

void psBounceOffFloor(ParticleSystem *ps){
	/*whack is broken.  It was intended to make the floor suddenly 
	  apply a force upward on all particles resting on it, but I
	  never took out the hacky method by which i was doing it.  When
	  whack-mode is off ("w" is not pressed) bouncing takes place with 
	  quite a bit of damping.
	*/
	int i;
	if(!whack){
	for(i=0; i<ps->numParticles; i++){
		if(ps->p[i].x[1]<=0){
			ps->p[i].v[0]*=0.9f;
			ps->p[i].v[1]*=-.2;
			ps->p[i].v[2]*=0.9f;
		}
	}
	}
	else{
	for(i=0; i<ps->numParticles; i++){
		if(ps->p[i].x[1]<=0){
			ps->p[i].v[1]+=-0.3f;
		}
	}
	whack=0;
	}
}

void psExplode(ParticleSystem * ps){
	/*This function just reinitializes all particles
	  and the particle system in order to cause an
	  "explosion".  Useful when changing modes to not 
	  have particles, for instance, stuck below the floor*/
	int i;
	ps->numParticles=0;
	for(i=0; i<ps->n; i++){
		pInit(ps,ps->p+i);
	}
}

psPathFollow(ParticleSystem * ps){
	/*This function causes the particles to follow a mathematical
	  function.  For speed purposes, each particle records the global
	  time at which it joined the path following stream. (either it was
	  born at that time, or pathfollowing was turned on at that time).  It
	  then increments this independently of the rest of the system.  The benefit
	  is that each particle is 1)faster to update, since no vector math is necessary
	  to figure out what direction the penalty force needs to be and 2)the particles
	  follow the path very closely.  The method besides this is the standard method
	  that was outlined in class on this subject (Penalty Forces).
    */
	int i;
	for(i=0; i<ps->numParticles; i++){
		//Here, the function
		ps->p[i].tpathjoin+=5;
		if(!ps->p[i].tpathjoin)
			ps->p[i].tpathjoin+=1;
		
		float dx = (ps->p[i].x[0]-3*sinf(ps->p[i].tpathjoin));
		float dy = (ps->p[i].x[1]-3*cosf(ps->p[i].tpathjoin));
		float dz = (ps->p[i].x[2]-3*sinf(ps->p[i].tpathjoin));
		
		float adx = sqrt(dx*dx);
		float ady = sqrt(dy*dy);
		float adz = sqrt(dz*dz);
		
		float dvx = (ps->p[i].v[0]-3*cosf(ps->p[i].tpathjoin));
		float dvy = (ps->p[i].v[1]+3*sinf(ps->p[i].tpathjoin));
		float dvz = (ps->p[i].v[2]-3*cosf(ps->p[i].tpathjoin));		
		ps->p[i].f[0] += (-(KS*adx + KD*(dvx*dx/adx)))*(dx/adx);
		ps->p[i].f[1] += (-(KS*ady + KD*(dvy*dy/ady)))*(dy/ady);
		ps->p[i].f[2] += (-(KS*adz + KD*(dvz*dz/adz)))*(dz/adz);
	}
	ps->x[0] = 3*sin(ps->t);
	ps->x[1] = 3*cos(ps->t);
	ps->x[2] = 3*sin(ps->t);
}

psDipoleAttract(ParticleSystem * ps){
	/*This function looks at a particle and its "friend".
	  Then it creates a force along a straight line between them.
	  It isn't that phenomenal an effect, but it looks sort of cool
	  to turn it on and off when the particle system is following
	  the path (see "pathfollow") above
    */
	int i;
	for(i=0; i<ps->numParticles; i++){
		float dx = ps->p[i].x[0]-ps->p[i].dual->x[0];
		float dy = ps->p[i].x[1]-ps->p[i].dual->x[1];
		float dz = ps->p[i].x[2]-ps->p[i].dual->x[2];

		ps->p[i].f[0] -= 2*dx; 
		ps->p[i].f[1] -= 2*dy; 
		ps->p[i].f[2] -= 2*dz;
	}
}



void psComputeForces(ParticleSystem *ps){
	int i;
	for(i=0; i<ps->numParticles; i++){
		ps->p[i].f[0] = 0;
		ps->p[i].f[1] = 0;
		ps->p[i].f[2] = 0;
	//Color of each particle is a linear interpolation between red and green based on its age, as a fraction
	//of the maximum allowable age
	float an = ((float)ps->p[i].age)/((float)EUTHENIZE);
	ps->p[i].c[0]=an;
	ps->p[i].c[1]=1-an;
	ps->p[i].c[2]=0;
	ps->p[i].c[3]=.30;
	}
	/*this is a list of potential forces that can affect
	  the P.S.  each has a global var that turns it on or off
	  see keyboard() for more info on which key triggers each*/
	if(gravity)	
		psApplyGravity(ps);
	if(floored)	
		psBounceOffFloor(ps);
	if(dipole)	
		psDipoleAttract(ps);
	if(pathf)	
		psPathFollow(ps);
	if(doExplode){
		doExplode=0;
		psExplode(ps);
	}
}


/*the following are vector manipulation functions i wrote.  They may not
  be optimal for speed or what have you, but they seem to be correct*/
void psGetScaleVector(ParticleSystem * ps, float * dst, float h){
	int i;
	for(i=0; i<STATE_LENGTH*ps->numParticles; i++){
		*(dst++)*=h;
	}
}

void psGetAddVectors(ParticleSystem * ps, float * dst, float * src){
	int i;
	for(i=0; i<STATE_LENGTH*ps->numParticles; i++){
		*(dst++)+=*(src++);
	}
}

void psKill(ParticleSystem *ps){
	/*kill all elderly particles in the system*/
	int i;
	for(i=0; i< ps->numParticles; i++){
		if((++(ps->p[i].age))>=EUTHENIZE){
			pInit(ps,ps->p+i);
			ps->numParticles--;
		}
	}
}

void psEulerSolver(ParticleSystem *ps,float h){
	//figure out whose turn it is to respawn at the emitter
	psKill(ps);
	
	//compute all the forces currently affecting the system
	psComputeForces(ps);
	
	//get the derivative of the current phase space vector (x[3],v[3])' = (v[3],a[3])
	psStateDeriv(ps,stateDeriv);
	
	//look ahead by one step of H, as prescribed by Euler
	psGetScaleVector(ps,stateDeriv,h);
	
	//get the current state of things in the system
	psGetParticleState(ps,state);
	
	//add the looked-ahead phase vector and the current step's vector
	psGetAddVectors(ps,state,stateDeriv);
	
	//update the particles
	psSetParticleState(ps,state);
	
	//of course, time increases monotonically forever =)
	ps->t+=h;
}

void psInit(ParticleSystem * ps){
	/*a pretty simple function to initialize a particle system*/
	int i;
	ps->n = MAX_NUM_P;
	ps->numParticles=0;
	ps->partRate=RATE;
	ps->t = 0;
	ps->x[0]=0;
	ps->x[1]=1;
	ps->x[2]=0;
	for(i=0; i<ps->n; i++){
		if(i%2==0)	
			ps->p[i].dual = ps->p+i+1;
		else
			ps->p[i].dual = ps->p+i-1;
	}

}


void init(){
	
	//I am using vertex/color arrays to manage the particle data
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	
	//I am using blending to make the particles not look like a mess when they pile up
	//(even though, since i'm not rendering back to front, it isn't *correct*, it looks better)
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

}


static void timer(int count){

		//Most of this is used to just manage time based particle release
		//and glutPostRedisplay-ing
		int i;
		//If we can fit more particles in here
		if(P->numParticles+P->partRate < P->n){
			//increase the number we have
			P->numParticles+=P->partRate;
			//we need to initialize all the particles we just introduced
			for(i=P->numParticles-P->partRate;i<P->numParticles;i++)
				pInit(P,P->p+i);
		}
		else
			P->numParticles=P->n;

		//Do the euler solver, with a smallish time step
		//this turns out to be of little consequence for any
		//of the effects that i've implemented, due to the way
		//i manage the time parameter of particles following a 
		//curve
		psEulerSolver(P,.05);
		
		//run "display()"
		glutPostRedisplay();
		
		//repost the timer function w/ a smallish delay
		glutTimerFunc(10,timer,0);
}



//==================================================================
//  Callback function called by GLUT when a key is pressed.
//  Parameters:
//  key - The key pressed.
//  x, y - Mouse position when the key is pressed.
//==================================================================
void keyboard (unsigned char key, int x, int y)
{
    switch (key)		
    {
    //----Exit
	case 'Q':
	case 'q':
        exit(0);	
        break;
	case 'x':
		changing=0;
	break;
	case 'y':
		changing=1;
	break;
	case 'z':
		changing=2;
	break;
	case 'j':
		P->x[changing]+=.1f;
	break;
	case 'k':
		P->x[changing]-=.1f;
	break;
	case 'e':
		doExplode=1;
	break;
    
	case 'w':
		whack=1;
	break;
	case 'h':
		R+=1.0;
	break;
	case 'l':
		R-=1.0;
	break;
//mode toggles:	
//These are fairly self explenatory:
	/*Press g to toggle Gravity
	  Press f to toggle floor bouncing (at z=0)
	  Press p to make the particles follow a mathematical function (x=sin(t) y=cos(t) z=(sin(t), as it happens)
	  Press d to make the particles pull toward their "friend" particle by a linear straight line force
	  Press h to rotate the camera left around the origin
	  Press l to rotate the camera right around the origin
    */
	case 'g':
		if(gravity)
			gravity=0;
		else
			gravity=1;
	break;
	case 'f':
		if(floored)
			floored=0;
		else
			floored=1;
	break;
	case 'p':
		if(pathf){
			pathf=0;
			P->x[0]=P->x[2]=0;
			P->x[1]=1;
		}
		else{
			P->x[0]=3*sin(P->t);
			P->x[1]=3*cos(P->t);
			P->x[2]=3*sin(P->t);
			pathf=1;
		}
	break;
	case 'd':
		if(dipole)
			dipole=0;
		else
			dipole=1;;
	break;
    default:
	break;
	}
}


//==================================================================
// Redraw the window
//==================================================================
void display (void)
{    
	int i;
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    
	gluLookAt(10 *cos(rad(R)) ,2  ,  10*sin(rad(R)),
		  0.0,0.0,0.0,
	          0.0,1.0,0.0);
	glPointSize(3.0f); 
	//Render vertex array as points
	glVertexPointer(3,GL_FLOAT,sizeof(Particle),P->p);
	glColorPointer(4,GL_FLOAT,sizeof(Particle),P->p->c);
	glDrawArrays(GL_POINTS,0,P->numParticles);
    
	glutSwapBuffers();
}

void reshape(int w, int h){
	if(w>h)
		glViewport((w-h)/2,0,h,h);
	else
		glViewport(0,(h-w)/2,w,w);

}



//==================================================================
// main()
//==================================================================
int main (int argc, char* argv[])
{
    srand(time(NULL));
    //----Glut initialization
	//glewInit();
    glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);       
    glutInitWindowSize(1024, 768);
    glutInitWindowPosition(180, 100);
    glutCreateWindow("Particle System");
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    init();

    //----Register callback functions
    	glutDisplayFunc(display);
    	//glutIdleFunc(idle);
	glutKeyboardFunc(keyboard);  
	glutReshapeFunc(reshape);
	glutTimerFunc(100,timer,0);

    //---- Set view parameters
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glMatrixMode (GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60,1,2,25);
//    glEnable(GL_DEPTH_TEST);
	
    //----Begin
	//Do all of those allocations.  I do them here, not on the stack
	//because w/ 10000 particles i got "stack overflows" so I wanted 
	//error checking during allocation
    if(!(P = (ParticleSystem *)malloc(sizeof(ParticleSystem)))){
	    printf("couldn't allocate particle system");
	    exit(0);
    }
    printf("allocating...\n");
    if(!(parts = (Particle *)malloc(sizeof(Particle) * MAX_NUM_P))){
	    printf("couldn't allocate particles ");
	    exit(0);
    }
	if(!(state = (float *) malloc(sizeof(float)*MAX_STATE_SIZE))){
		printf("couldn't allocate state vector\n");
		exit(0);
	}
	if(!(stateDeriv = (float *) malloc(sizeof(float)*MAX_STATE_SIZE))){
		printf("couldn't allocate state derivitive vector\n");
		exit(0);
	}
    printf("done!\n");
    
	//Point the global particle system's particle pointer at "parts"
    P->p = parts;
	//initialize
    psInit(P);
	//go into glut-ville
    glutMainLoop();
    return 0;
}


